/*
 *  linux/fs/supermount/shadow.c
 *
 *  Original version:
 *      Copyright (C) 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 * $Id: shadow.c,v 1.22 2000/09/04 16:06:33 root Exp $
 *
 *  Manage shadow inodes.
 */

#include <asm/segment.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/mm.h>

static int  supermount_shadow_get_write_access (struct super_block *);
static void supermount_shadow_put_write_access (struct super_block *);
static int supermount_check_writeable(struct super_block *sb);
static void supermount_update_atime(struct inode *inode);

struct inode_shadow_operations supermount_shadow_iops = {
	NULL /*supermount_shadow_get_write_access*/,
	NULL /*supermount_shadow_put_write_access*/,
	supermount_check_writeable,
	NULL, /* set_attr */
	supermount_update_atime
};

static int supermount_shadow_get_write_access (struct super_block *sb)
{
	int rc = 0;
//	supermount_dump_inode(inode);
	supermount_debug("supermount_shadow_get_write_access\n");
	rc = supermount_get_write_access(sb);
	if (!rc)
		supermount_iopen(sb);
	return rc;
}

static void supermount_shadow_put_write_access (struct super_block *sb)
{
	supermount_debug("supermount_shadow_put_write_access\n");
//	supermount_dump_inode(inode);
	mark_subfs_dirty(sb);
	supermount_put_write_access(sb);
	supermount_iclose(sb);
}

static int supermount_check_writeable(struct super_block *sb)
{
//	supermount_dump_inode(inode);
	return !(sb->s_flags & MS_RDONLY);
}

struct dentry *get_superfs_dentry_by_inode(struct inode *inode);

static void supermount_update_atime(struct inode *inode)
{
	struct dentry *dentry = get_superfs_dentry_by_inode(inode);

	if (inode->i_shadow == NULL)
		goto dput_out;

	supermount_get_write_access(inode->i_shadow);
	mark_inode_dirty(inode);
	mark_subfs_dirty(inode->i_shadow);
	if (dentry != NULL && dentry->d_inode != NULL)
		dentry->d_inode->i_atime = inode->i_atime;
	supermount_put_write_access(inode->i_shadow);

dput_out:
	if (dentry)
		dput(dentry);
	supermount_go_inactive(inode->i_shadow);
}
